#include "p4.h"
#include <signal.h>
#include <stdio.h>
#include <math.h>

/* #define VERBOSE */

#define USERBUF

#if defined(TRANSPUTER)
#define TRANSFER 100000
#else
#define TRANSFER 1000
#endif

#define MEG (1024*1024)

int master();
int slave();

int main(int argc,char* argv[])
{
  p4_initenv(&argc,argv);
  if (p4_get_my_id() == 0)
      p4_create_procgroup();

#if defined(VERBOSE)
  printf("in main as %d\n", p4_get_my_id());
#endif

  switch (p4_get_my_id()) {
  case 0:
    master();
    break;
  case 1:
    break;
  case 2:
    slave();
    break;
  }
  p4_wait_for_end();
}

int repeats;
int size;

#define INIT 1
#define TIME 2

#define FIRST 0
#define LAST 40

int master() {
  int i,j;
  char* msg;
  char* reply;
  char* rep;
  int* got;
  int len;
  p4_usc_time_t start;
  p4_usc_time_t end;
  p4_usc_time_t elapsed;
  int from, type; 
  int bytes;
  double band, ping;
  int on = 1;
  int* int_msg;

  int_msg = (int*)p4_msg_alloc(sizeof(int));

#if defined(VERBOSE)
      printf("<0:sending %d>\n",sizeof(int));
#endif
  
  *int_msg = on;
  p4_sendr(INIT,2,int_msg,sizeof(int));
#if defined(VERBOSE)
      printf("<0:sent %d>\n",sizeof(int));
#endif

  size = (int) pow(10.0, 1.0 + LAST/10.0);
#if defined(VERBOSE)
      printf("<0:sending %d>\n",sizeof(int));
#endif
  *int_msg = size;
  p4_sendr(INIT,2,int_msg,sizeof(int));
#if defined(VERBOSE)
      printf("<0:sent %d>\n",sizeof(int));
#endif

/*  msg = p4_malloc(size); !!!!!!!!!! */
  msg = p4_msg_alloc(size);
#if defined(USERBUF)
  rep = reply = p4_msg_alloc(size);
#endif

  for (j=FIRST; j<=LAST; j++) {
    size = (int) pow(10.0, 1.0+j/10.0);
    repeats = TRANSFER / size;

    printf("%d %d ", size, repeats);

#if defined(VERBOSE)
      printf("<0:sending %d>\n",sizeof(int));
#endif

    *int_msg = repeats;
    p4_sendr(INIT,2,int_msg,sizeof(int));
#if defined(VERBOSE)
      printf("<0:sent %d>\n",sizeof(int));
#endif
#if defined(VERBOSE)
      printf("<0:sending %d>\n",sizeof(int));
#endif

    *int_msg = size;
    p4_sendr(INIT,2,int_msg,sizeof(int));
#if defined(VERBOSE)
      printf("<0:sent %d>\n",sizeof(int));
#endif

    start = p4_ustimer();
    
    from = 2;
    type = TIME;
    len = size;

    for (i=0; i<repeats; i++) {
#if defined(VERBOSE)
      printf("<0:sending %d>\n",size);
#endif
      p4_sendb(TIME,2,msg,size);
#if defined(VERBOSE)
      printf("<0:sent %d>\n",size);
#endif
#if !defined(USERBUF)
      reply = 0;
#else
      reply = rep;
#endif
      from = 2;
      type = TIME;
      len = size;
#if defined(VERBOSE)
      printf("<0:recieving %d>\n", len);
#endif
      p4_recv(&type,&from,(char**)&reply,&len);
#if defined(VERBOSE)
      printf("<0:recieved %d>\n",len);
#endif
#if !defined(USERBUF)
      p4_msg_free(reply);
#endif
    }
    end = p4_ustimer();
    
    elapsed = end - start;
    bytes = size * repeats * 2;
    band = ((double) bytes / MEG) / ( (double)elapsed / 1000000 );
    ping = (double) elapsed / (double) repeats;

#if defined(VERBOSE)
    printf("elapsed time=%d us\n", elapsed);
    printf("repeats = %d\n", repeats);
    printf("amount of data=%d b, %d Kb, %d Mb\n",
	   bytes, bytes / 1024, bytes / MEG);
    printf("bandwidth = %lf Mb/s\n", band);
#endif

    printf("%lf Mb/s %lf us\n", band, ping);

  }

#if defined(USERBUF)
  p4_msg_free(reply);
#endif    
  exit(0);
}

int slave() {
  int i;
  char* msg;
  char* reply;
  char* rep;
  int len;
  int* int_msg;
  int* imsg;
  int from, type;
  int on;

#if defined(VERBOSE)
  printf("in slave as %d\n", p4_get_my_id());
#endif

  from = 0;
  type = INIT;
  len = sizeof(int);
  imsg = int_msg = (int*)p4_msg_alloc(sizeof(int));
#if defined(VERBOSE)
      printf("<1:recieving %d>\n", len);
#endif
  p4_recv(&type,&from,&int_msg,&len);
#if defined(VERBOSE)
      printf("<1:recieved %d>\n", len);
#endif
  on = *int_msg;

  from = 0;
  type = INIT;
  len = sizeof(int);
  int_msg = imsg;
#if defined(VERBOSE)
      printf("<1:recieving %d>\n", len);
#endif
  p4_recv(&type,&from,&int_msg,&len);
#if defined(VERBOSE)
      printf("<1:recieved %d>\n", len);
#endif
  size = *int_msg;

  msg = p4_msg_alloc(size);
#if defined(USERBUF)
  rep = reply = p4_msg_alloc(size);
#endif

  while (1) {

#if defined(VERBOSE)
    printf("slave loop\n");
#endif

    from = 0;
    type = INIT;
    len = sizeof(int);
#if defined(VERBOSE)
      printf("<1:recieving %d>\n", len);
#endif
    int_msg = imsg;
    p4_recv(&type,&from,(char**)&int_msg,&len);
#if defined(VERBOSE)
      printf("<1:recieved %d>\n", len);
#endif
    repeats = *int_msg;
#if defined(VERBOSE)
    printf("repeats = %d\n", repeats);
#endif

    from = 0;
    type = INIT;
    len = sizeof(int);
#if defined(VERBOSE)
      printf("<1:recieving %d>\n", len);
#endif
    int_msg = imsg;
    p4_recv(&type,&from,&int_msg,&len);
#if defined(VERBOSE)
      printf("<1:recieved %d>\n", len);
#endif
    size = *int_msg;
#if defined(VERBOSE)
    printf("size = %d\n", size);
#endif

#if defined(VERBOSE)
    printf("slave got repeats = %d, size = %d\n", repeats, size);
#endif

    for (i=0; i<repeats; i++) {
      from = 0;
      type = TIME;
      len = size;
#if !defined(USERBUF)
      reply = 0;
#else
      reply = rep;
#endif
#if defined(VERBOSE)
      printf("<1:recieving %d>\n", len);
#endif
      p4_recv(&type,&from,&reply,&len);
#if defined(VERBOSE)
      printf("<1:recieved %d>\n",len);
#endif
#if !defined(USERBUF)
      p4_msg_free(reply);
#endif
#if defined(VERBOSE)
      printf("<1:sending %d>\n",size);
#endif
      p4_sendb(TIME,0,msg,size);
#if defined(VERBOSE)
      printf("<1:sent %d>\n",size);
#endif
    }
  }
}
