/**********************************************************************\
*                                                                      *
*                  Copyright (c) Exeter University,		       *
*                  Department of Computer Science,                     *
*                 (aba@uk.ac.ex.dcs) Adam Back, 1994                   *
*                                                                      *
*  GNU GPL (General Public License) Applies.                           *
*                                                                      *
\**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <rpc/rpc.h>
#include <netinet/in.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include "protocol.h"
#include "xdr.h"
#include "macros.h"
#include "sockxdr.h"
#include "common.h"

typedef struct Semaphore Semaphore;

struct Semaphore {
  int sid;
  int id;
};

#define DEBUG
#ifdef DEBUG
int debug = 0;
char line[ 100 ];
#endif

#define BUF_SIZE 4096

int sid;
int sem_limit;

void semcall( Semaphore *s, int op ) {
  struct sembuf sb;
  sb.sem_num = s->id;
  sb.sem_op = op;
  sb.sem_flg = 0;
  if ( semop( s->sid, &sb, 1 ) == -1 ) {
    perror( "semaphor operation failed" );
    fprintf( stderr, "errno = %d\n", errno );
    exit( 1 );
  }
}

void SemWait( Semaphore *s ) {
  semcall( s, -1 );
}

void SemSignal( Semaphore *s ) {
  semcall( s, 1 );
}

void SemSetAlloc( int n ) {
  static int serial = 0;
  if ( serial > 0 ) {
    fprintf( stderr, "SemSetAlloc called twice\n" );
    exit( 1 );
  }
  sem_limit = n;
  sid = semget( (key_t)getpid(), n, 0600 | IPC_CREAT );
  if ( sid == -1 ) {
    perror( "semaphor alloc failure" );
    fprintf( stderr, "errno = %d\n", errno );
    exit( 1 );
  }
  serial++;
}

void SemInit( Semaphore *s, int val ) {
  static int num = 0;
  if ( num > sem_limit ) {
    fprintf( stderr, "out of semaphores\n" );
    exit( 1 );
  }
  s->id = num;
  s->sid = sid;
  semcall( s, val );
  num++;
}

void KillIPC( int mid ) {
  semctl( mid, IPC_RMID, 0 );
}

void SemKill( Semaphore *s ) {
  KillIPC( s->sid );
}

void ChanOutChar( int fd, char ch )
{
  int ecode;

  do {
    ecode = write( fd, &ch, 1 );
    if ( ecode == -1 ) {
      perror( "ChanOutChar" );
      fprintf( stderr, "errno = %d\n", errno );
      exit(1);
    }
  } while ( ecode != 1 );
}

void ChanOut( int fd, void* buf, int size ){
  int ecode, sent = 0;
  do {
    ecode = write( fd, buf, size );
    if ( ecode > 0 ) {
      sent += ecode;
      buf += ecode;
    }
  } while ( sent < size );
}

void ChanIn( int fd, void* buf, int size ){
  int ecode, recvd = 0;
  do {
    ecode = read( fd, buf, size );
    if ( ecode > 0 ) {
      recvd += ecode;
      buf += ecode;
    }
  } while ( recvd < size );
}

int ChanInInt( int fd ){
  int i;
  char *buf = (char*)&i;
  int ecode, recvd = 0;
  do {
    ecode = read( fd, buf, sizeof( int ) );
    if ( ecode > 0 ) {
      recvd += ecode;
      buf += ecode;
    }
  } while ( recvd < sizeof( int ) );
  buf = (char*)&i;
  return (long)buf[3] << 24 | (long)buf[2] << 16 | 
         (long)buf[1] << 8 | (long)buf[0];
}

void ChanOutInt( int fd, int val ){
  char *buf = (char*) &val;
  int ecode, sent = 0;
  val = (long)buf[3] << 24 | (long)buf[2] << 16 | 
         (long)buf[1] << 8 | (long)buf[0];
  do {
    ecode = write( fd, buf, sizeof( int ) );
    if ( ecode > 0 ) {
      sent += ecode;
      buf += ecode;
    }
  } while ( sent < sizeof( int ) );
}

char ChanInChar( fd )
int fd;
{
  int ecode;
  char ch;

  do {
    ecode = read( fd, &ch, 1 );
    if ( ecode == -1 ) {
      perror( "ChanInChar" );
      fprintf( stderr, "errno = %d\n", errno );
      exit(1);
    }
  } while ( ecode != 1 );
  return ch;
}

char ChanInC( d )
FILE *d;
{
  int ecode;
  int ch = 0;
  int reset = 1;

  do {
    ch = 0;
    if ( feof( d ) ) {
      fprintf( stderr, "end of file\n" );
    } 
    ch = fgetc( d );
    if ( ch == EOF ) clearerr( d );
  } while ( ch == EOF );
  return ch;
}

void serve(void);
void problem(void);

void do_socket(u_int);
void do_listen(u_int);
void do_accept(u_int);
void do_select(u_int);
void do_connect(u_int);
void do_bind(u_int);
void do_read(u_int);
void do_write(u_int);
void do_close(u_int);
void do_recv(u_int);
void do_recvfrom(u_int);
void do_send(u_int);
void do_sendto(u_int);
void do_getsockname(u_int);
void do_gethostbyname(u_int);
void do_getpeername(u_int);
void do_shutdown(u_int);
void do_fcntl(u_int);
void do_chdir(u_int);
void do_ss_puts(u_int);

int sstep = 0, lline = 0, dxdr = 0, start = 0;

#define DO(x) {char test[100]; if (dxdr) fprintf(stderr,"SERVER:%s\n",#x); \
if (sstep) gets(test); \
if (x == 0) return; }

/*************************************************************\
*                                                             *
* Variables for transputer device which is passed as          *
* parameter.                                                  *
*                                                             *
\*************************************************************/

XDR xdr_out, xdr_in;     /* eXternal Data Rep. streams see xdr(3n)      */

int serverIn, serverOut; /* file descriptors connected to device        */

FILE *serverInStream;    /* stream version of serverIn as XDR expects a */
                         /* stream not a file descriptor                */

/************************************************************/

int dbg=0;

void synch( int fd ){
  char c;
  c=1;
  while (c!=0) {
    c = ChanInChar( fd );
  }
  while (c==0) {
    putchar('.'); fflush( stdout );
    c = ChanInChar( fd );
  }
  if ( c != 1 ) {
    printf( "warning: synchronization error\n" );
  }
}

void init( char* device ){
  int i,xdrlong;
  char *startenv;
  FILE *d;
  long test=charPACK('A','d','a','m');
  char c;
  int fd;
  char realmesg[100];
  char **mesg=(char**)malloc(sizeof(char*));
  
  *mesg = realmesg;

  if (getenv("DEBUG")) dbg = 1;
  if (getenv("SSTEP")) sstep = 1;
  if (getenv("LINES")) lline = 1;
  if (getenv("XDR")) dxdr = 1;
  if ((startenv=getenv("START"))) start=atoi(startenv);

  if ( ( fd = open( device, O_RDWR ) ) == -1 ) {
    fprintf(stderr,"Could not open device %s\n",device);
    exit(1);
  }

/*****************************************************************************
 * MCP1000 hack                                                              *
 *                                                                           *
 * The transtech iserver does this ioctl, which we do not have documentation *
 * for so we copy it verbatim from the trace ( see man trace(1) ) output.    *
 *                                                                           *
 *****************************************************************************/

#define MCP1000_MAGIC 0x20006e00

  if ( ioctl( fd, MCP1000_MAGIC, 0 ) == -1 ) {
    perror("ioctl failed on stdin");
    fprintf( stderr, "errno = %d\n", errno );
    exit(1);
  }

/****************************************************************************/

  serverIn = serverOut = fd;

  if ( ( serverInStream = fdopen( fd, "rw" ) ) == 0 ) {
    perror( "could not convert fd" );
    fprintf( stderr, "errno = %d\n", errno );
    exit( 1 );
  }

  synch( fd );
}

int id;

struct info {
  int len;
  int processor;
  int process;
} *Info, MyInfo;

void server( void );

struct id_info {
  int unixid;
  Semaphore lock;
};

Semaphore got_write_perm, got_read_perm;
struct id_info *ids;
int nids, mid;
char *buf;

#define DONE() SemSignal( &got_read_perm )

void cleanup();

char *disp;
FILE *dispfd;

void main(int argc, char *argv[] ){
  int size;
  FILE *fd;
  int i;

#ifdef DEBUG
  if ( argc > 1 && argv[1][0] == '-' && argv[1][1] == 'd' ) {
    debug = 1;
    argv++;
    argc--;
    if ( argc > 2 ) {
      disp = argv[ 1 ];
      argv++;
      argc--;
      dispfd = fopen( disp, "w" );
      if ( dispfd == NULL ) {
	perror( "open display:" );
	exit( 1 );
      }
    } else {
      dispfd = stdout;
    }
  }
#endif
  init( argv[ 1 ] );
/*   if ( argc > 2 ) nids = atoi( argv[ 2 ] ); */
  nids = ChanInInt( serverIn ) * MAX_SOCKET_PROCS;
  printf( "nids = %d\n", nids );
  ids = ( struct id_info* ) malloc( sizeof( struct id_info ) * nids );
  if ( ids == 0 ) { fprintf( stderr, "Out of memory\n" ); exit( 1 ); }
  nids = 50;
  SemSetAlloc( nids + 2 );
  SemInit( &got_write_perm, 1 );
  SemInit( &got_read_perm, 1 );
  for ( i = 0; i < nids; i++ ) {
    ids[i].unixid = 0;
    SemInit( &ids[ i ].lock, 0 );
  }
  if ( ( mid = shmget( getpid(), sizeof( struct info ), IPC_CREAT|0600 ) ) == -1 ) {
    perror( "sserver" );
    fprintf( stderr, "errno = %d\n", errno );
    exit( 1 );
  }

  if ( ( Info = (struct info*)shmat( mid, 0, 0 ) ) == (struct info*)-1 ){
    perror( "sserver" );
    fprintf( stderr, "errno = %d\n", errno );
    exit( 1 );
  }

  xdrstdio_create( &xdr_in, serverInStream, XDR_DECODE );

  while ( 1 ) {
    SemWait( &got_read_perm );

    printf( "main server regained control\n" );

    Info->processor = ChanInInt( serverIn );
    Info->process = ChanInInt( serverIn );
    Info->len = ChanInInt( serverIn );
    printf( "Info = { processor = %d, process = %d, len = %d }\n", 
	    Info->processor, Info->process, Info->len );
    id = Info->processor * MAX_SOCKET_PROCS + Info->process;
    if ( ids[ id ].unixid == 0 ){
      ids[ id ].unixid = fork();
      if ( ids[ id ].unixid == 0 ){
#ifdef DEBUG
	printf( "child blocked\n" ); fflush( stdout );
	printf( "resume child" );
	gets( line );
	printf( "child released\n" ); fflush( stdout );
#endif
	if ( ( Info = (struct info*)shmat( mid, 0, 0 ) ) == (struct info*)-1 ){
	  perror( "sserver" );
	  fprintf( stderr, "errno = %d\n", errno );
	  exit( 1 );
	}
	MyInfo = *Info;
	buf = (char*) malloc( BUF_SIZE);
	if ( buf == 0 ) {
	  fprintf(stderr, "Out of memory\n");
	  exit( 1 );
	}
	xdrmem_create( &xdr_out, buf, BUF_SIZE, XDR_ENCODE );
	while ( 1 ) {
	  SemWait( &ids[ id ].lock );
	  server();
	}
      } else {
	printf( "parent forked child %d\n", ids[ id ].unixid );
      }
    }
    SemSignal( &ids[ id ].lock );    /* release its server */
  }
}

void cleanup() {
  int i;
  for ( i = 0; i < nids; i++ ) {
    if ( ids[ i ].unixid ) 
      kill( ids[ i ].unixid, SIGQUIT );
  }
  sleep( 1 );
  for ( i = 0; i < nids; i++ ) {
    if ( ids[ i ].unixid ) 
      kill( ids[ i ].unixid, SIGSTOP );
  }
/* free shared memory segment */
  KillIPC( mid );
/* free up all semaphores associated with sid */
  SemKill( &got_read_perm );
  exit(0);
}

void server(){
  long command = 0;
  printf( "server started\n" );
  if (!xdr_long(&xdr_in, &command)) problem();
  switch (CHAR(1,command)){
  case SERV_socket        : R(do_socket(command));        break;
  case SERV_listen        : R(do_listen(command));        break;
  case SERV_accept        : R(do_accept(command));        break;
  case SERV_select        : R(do_select(command));        break;
  case SERV_connect       : R(do_connect(command));       break;
  case SERV_bind          : R(do_bind(command));          break;
  case SERV_read          : R(do_read(command));          break;
  case SERV_write         : R(do_write(command));         break;
  case SERV_close         : R(do_close(command));         break;
  case SERV_getsockname   : R(do_getsockname(command));   break;
  case SERV_gethostbyname : R(do_gethostbyname(command)); break;
  case SERV_getpeername   : R(do_getpeername(command));   break;
  case SERV_shutdown      : R(do_shutdown(command));      break;
  case SERV_recv          : R(do_recv(command));          break;
  case SERV_recvfrom      : R(do_recvfrom(command));      break;
  case SERV_send          : R(do_send(command));          break;
  case SERV_sendto        : R(do_sendto(command));        break;
  case SERV_fcntl         : R(do_fcntl(command));         break;
  case SERV_chdir         : R(do_chdir(command));         break;
  case SERV_ss_puts       : R(do_ss_puts(command));         break;
  case SERV_reset         : kill( getppid(), SIGUSR1 );
                            exit( 0 );
  default:;
  }
  SemWait( &got_write_perm );
  ChanOutInt( serverOut, MyInfo.processor );
  ChanOutInt( serverOut, MyInfo.process );
  ChanOutInt( serverOut, xdr_getpos( &xdr_out ) );
  ChanOut( serverOut, buf, xdr_getpos( &xdr_out ) );
  xdr_setpos( &xdr_out, 0 );
  SemSignal( &got_write_perm );
}

void problem(void){
  fprintf(stderr,"Server problem: transputer not responding\n");
  exit(1);
}

void do_socket(u_int command){
  int result;
  int domain =   CHAR(2,command);
  int type =     CHAR(3,command);
  int protocol = CHAR(4,command);

  DONE();

  fprintf(stderr,"socket(%d,%d,%d)=",domain,type,protocol);
  result = socket(domain,type,protocol);
  fprintf(stderr,"%d\n",result);
  DO(xdr_long(&xdr_out,&result));
  if (result == -1) xdr_long(&xdr_out,&errno);
}

void do_listen(u_int command){
  int result;
  int backlog = CHAR(2,command);
  int fd =      SHORT(2,command);

  DONE();

L  fprintf(stderr,"listen(%d,%d)=",fd,backlog); 
L  result = listen(fd, backlog);
L  fprintf(stderr,"%d\n",result);

L  DO(xdr_long(&xdr_out,&result));
  fprintf(stderr,"sent result\n");
L  if (result == -1) xdr_long(&xdr_out,&errno);
}

void do_accept(u_int command){
  char temp[256];
  int result, fd = SHORT(2,command);
  struct sockaddr *from=(struct sockaddr *)temp;
  int fromlen=256;
  
  DONE();

  fprintf(stderr,"accept(%d,&,&)=",fd); 
  if (CHAR(2,command)) from = 0;
  result = accept(fd, from, &fromlen);
  fprintf(stderr,"%d,len=%d\n",result,fromlen);

  DO(xdr_long(&xdr_out,&result));
  if (result != -1 && !CHAR(2,command)){
    DO(xdr_sockaddr(&xdr_out,from,&fromlen));
  }
  if (result == -1) xdr_long(&xdr_out,&errno);
}

void do_select(u_int command){
  int result, nfds=SHORT(2,command);
  char flags=CHAR(2,command);
  fd_set readfds, writefds, exceptfds;
  struct timeval timeout;
  int i,nlongs = howmany(nfds,32);

  fprintf(stderr,"Bits set:\"");
  if (BIT(1,flags)) fprintf(stderr,"r");
  if (BIT(2,flags)) fprintf(stderr,"w");
  if (BIT(3,flags)) fprintf(stderr,"e");
  fprintf(stderr,"\"\n");

  if (!BIT(1,flags)) 
    for (i=0; i<nlongs; i++)
      DO(xdr_long(&xdr_in, &readfds.fds_bits[i]));
  if (!BIT(2,flags))
    for (i=0; i<nlongs; i++)
      DO(xdr_long(&xdr_in, &writefds.fds_bits[i]));
  if (!BIT(3,flags)) 
    for (i=0; i<nlongs; i++) 
      DO(xdr_long(&xdr_in, &writefds.fds_bits[i]));
  if (!BIT(4,flags)) {
    DO(xdr_long(&xdr_in, &timeout.tv_sec));
    DO(xdr_long(&xdr_in, &timeout.tv_usec));
  }

  DONE();

  result = select(nfds,
		  (BIT(1,flags) ? 0 : &readfds),
		  (BIT(2,flags) ? 0 : &writefds),
		  (BIT(3,flags) ? 0 : &exceptfds),
		  (BIT(4,flags) ? 0 : &timeout));
  DO(xdr_long(&xdr_out, &result));
  if (BIT(1,flags))
    for (i=0; i<nlongs; i++)
      DO(xdr_long(&xdr_out, &readfds.fds_bits[i]));
  if (BIT(2,flags))
    for (i=0; i<nlongs; i++)
      DO(xdr_long(&xdr_out, &writefds.fds_bits[i]));
  if (result == -1) xdr_long(&xdr_out,&errno);
}

void do_connect(u_int command){
  long templong;
  char temp[256];
  struct sockaddr *to=(struct sockaddr *)temp;
  int result, tolen, fd = SHORT(2,command);
  if (!CHAR(2,command)){
    DO(xdr_sockaddr(&xdr_in,to,&tolen));
  } else
    to=0;

  DONE();

  result = connect(fd,to,tolen);
  DO(xdr_long(&xdr_out,&result));
  if (result == -1) xdr_long(&xdr_out,&errno);
}

void do_bind(u_int command){
  char temp[256];
  int result, namelen, fd = SHORT(2,command);
  struct sockaddr *name=(struct sockaddr *)temp;
  DO(xdr_sockaddr(&xdr_in,name,&namelen));

  DONE();

  fprintf(stderr,"bind(%d,*,*)=",fd);
  result = bind(fd,name,namelen);
  fprintf(stderr,"%d\n",result);
  DO(xdr_long(&xdr_out,&result));
  if (result == -1) xdr_long(&xdr_out,&errno);
}

char *rbuf = 0;
u_int bufsize = 0;

void grow(int reqd){
  if (bufsize < reqd) {
    if (rbuf != 0) free(rbuf);
    rbuf = (char*)malloc(reqd);
    bufsize = reqd;
  }
}

void do_read(u_int command){
  int result, fd = SHORT(2,command);
  u_int nbytes;
  DO(xdr_long(&xdr_in,&nbytes));

  DONE();

  grow(nbytes);
  fprintf(stderr,"read(%d,&,%d)=",fd,nbytes);
  result = read(fd,rbuf,nbytes);
  fprintf(stderr,"%d\n",result);
  DO(xdr_long(&xdr_out,&result));
  if (result > 0) {
    if (xdr_opaque(&xdr_out,rbuf,result) ==  0) return;
  } else if (result == -1)
    xdr_long(&xdr_out,&errno);
}

void do_write(u_int command){
  int result, fd = SHORT(2,command);
  u_int nbytes;
  DO(xdr_long(&xdr_in,&nbytes));
  grow(nbytes);
  DO(xdr_opaque(&xdr_in,rbuf,nbytes));

  DONE();

  result = write(fd, rbuf,nbytes); 
  DO(xdr_long(&xdr_out,&result));
  if (result == -1)
    xdr_long(&xdr_out,&errno);
}

void do_close(u_int command){
  int result, fd = SHORT(2,command);
  result = close(fd);

  DONE();

  DO(xdr_long(&xdr_out,&result));
  if (result == -1)
    xdr_long(&xdr_out,&errno);
}

void do_recv(u_int command){
  int result, fd = SHORT(2,command);
  int len, flags;
  DO(xdr_long(&xdr_in,&len));
  DO(xdr_long(&xdr_in,&flags));

  DONE();

  grow(len);
  result = recv(fd,rbuf,len,flags);
  DO(xdr_long(&xdr_out,&result));
  if (result > 0) {
    if (xdr_opaque(&xdr_out,rbuf,result) ==  0) return;
  } else if (result == -1)
    xdr_long(&xdr_out,&errno);
}
   
void do_recvfrom(u_int command){
  char temp[256];
  int result, fd = SHORT(2,command);
  int len, flags, fromlen;
  struct sockaddr *from = (struct sockaddr*) temp;
  DO(xdr_long(&xdr_in,&len));
  DO(xdr_long(&xdr_in,&flags));

  DONE();

  grow(len);
  result = recvfrom(fd,rbuf,len,flags,from,&fromlen);
  DO(xdr_long(&xdr_out,&result));
  if (result > 0) {
    if (!CHAR(2,command))
      DO(xdr_sockaddr(&xdr_out,from,&fromlen));
    DO(xdr_opaque(&xdr_out,rbuf,result));
  } else if (result == -1)
    xdr_long(&xdr_out,&errno);
}

void do_send(u_int command){
  int result, fd = SHORT(2,command);
  int len, flags;
  DO(xdr_long(&xdr_in,&len));
  DO(xdr_long(&xdr_in,&flags));

  DONE();

  grow(len);
  DO(xdr_opaque(&xdr_in,rbuf,len));
  result = send(fd,rbuf,len,flags);
  DO(xdr_long(&xdr_out,&result));
  if (result == -1)
    xdr_long(&xdr_out,&errno);
}

void do_sendto(u_int command){
  char temp[256];
  struct sockaddr *to = (struct sockaddr*) temp;
  int tolen;
  int result, fd = SHORT(2,command);
  int len, flags;
  DO(xdr_long(&xdr_in,&len));
  DO(xdr_long(&xdr_in,&flags));
  DO(xdr_sockaddr(&xdr_in,to,&tolen));
  grow(len);
  DO(xdr_opaque(&xdr_in,rbuf,len));

  DONE();

  fprintf(stderr,"sendto(%d,\"%s\",%d,%d,{%d,%s})=",fd,rbuf,len,flags,
	 to->sa_family,to->sa_data);
  result = sendto(fd,rbuf,len,flags,to,tolen);
  fprintf(stderr,"%d\n",result);
  DO(xdr_long(&xdr_out,&result));
  if (result == -1)
    xdr_long(&xdr_out,&errno);
}

void do_getsockname(u_int command){
  char temp[256];
  int result, fd = SHORT(2,command);
  struct sockaddr *from=(struct sockaddr *)temp;
  int fromlen=256;
  
  DONE();

  fprintf(stderr,"getsockname(%d,&,&)=",fd);
  result = getsockname(fd, from, &fromlen);
  fprintf(stderr,"%d\n",result);
  DO(xdr_long(&xdr_out,&result));
  if (result != -1){
    DO(xdr_sockaddr(&xdr_out,from,&fromlen));
  }
  if (result == -1) xdr_long(&xdr_out,&errno);
}

void do_gethostbyname(u_int command){
  int len=SHORT(2,command);
  struct hostent *entry;
  char *name=(char*)malloc(len);
  DO(xdr_opaque(&xdr_in,name,len));

  DONE();

  entry=gethostbyname(name);
  xdr_hostent(&xdr_out,entry);
}

void do_getpeername(u_int command){
  char temp[256];
  int result, fd = SHORT(2,command), namelen = 256;
  struct sockaddr *name=(struct sockaddr *)temp;

  DONE();

  result = getpeername(fd, name, &namelen);
  DO(xdr_long(&xdr_out,&result));
  if (result != -1){
    DO(xdr_sockaddr(&xdr_out,name,&namelen));
  } else
    xdr_long(&xdr_out,&errno);
}

void do_shutdown(u_int command){
  int result, fd = SHORT(2,command);
  int how=CHAR(2,command);

  DONE();

  result = shutdown(fd,how);
  DO(xdr_long(&xdr_out,&result));
  if (result == -1)
    xdr_long(&xdr_out,&errno);
}

void do_fcntl(u_int command){
  int result, fd = SHORT(2,command);
  int cmd, arg;
  DO(xdr_long(&xdr_in,&cmd));
  DO(xdr_long(&xdr_in,&arg));

  DONE();

  result = fcntl(fd,cmd,arg);
  DO(xdr_long(&xdr_out,&result));
  if (result == -1)
    xdr_long(&xdr_out,&errno);
}

void do_chdir(u_int command){
  int result;
  char space[2048], *path=space;
  int cmd, arg;
  DO(xdr_string(&xdr_in,&path,2048));

  DONE();

  result = chdir(path);
  DO(xdr_long(&xdr_out,&result));
  if (result == -1)
    xdr_long(&xdr_out,&errno);
}

void do_ss_puts(u_int command){
  int result;
  char space[2048], *string=space;
  int cmd, arg;
  DO(xdr_string(&xdr_in,&string,2048));

  DONE();

  result = fputs( string, dispfd );
  fflush( dispfd );

  DO(xdr_long(&xdr_out,&result));
  if (result == -1)
    xdr_long(&xdr_out,&errno);
}
