/**********************************************************************\
*                                                                      *
*                  Copyright (c) Exeter University,		       *
*                  Department of Computer Science,                     *
*                 (aba@uk.ac.ex.dcs) Adam Back, 1994                   *
*                                                                      *
*  GNU GPL (General Public License) Applies.                           *
*                                                                      *
\**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <channel.h>
#include <semaphor.h>
#include <misc.h>
#include <process.h>
#include "t_fork.h"
#include "t_socket.h"
#include "t_xdr.h"
#include "t_sockxdr.h"
#include "t_protocol.h"
#include "t_macros.h"
#include "t_common.h"
#include "t_cfs.h"

#define BUF_SIZE 1024

PUBLIC int sstep = 0, lline = 0, dxdr = 0, start = 0;
PRIVATE XDR xdr_in,xdr_out;

PRIVATE void cleanup(void){
  long command=charPACK(SERV_reset,_,_,_);
  xdr_long(&xdr_out,&command);
}

/* PRIVATE int socketN; /* unused what was it for? */
PRIVATE Channel *synchChan;
PRIVATE Channel *socketIn, *socketOut;
PRIVATE Semaphore *socketOutLock;

void Demuxer( Process *p, Channel *synchChan, Channel *socketIn ) {
  int who, len;
  while ( 1 ) {
    who = ChanInInt( socketIn );
    len = ChanInInt( socketIn );
    xdr_setpos( &xdr_in, 0 );
    ChanIn( socketIn, xdr_in.x_base, len );
    ChanOutChar( &synchChan[ who ], 'w' );
    ChanOutChar( &synchChan[ who ], 'w' );
  }
}

int RegisterFork_with_SocketLib( void ){
  void *read_buf, *write_buf;
  read_buf = (void*) malloc( BUF_SIZE );
  write_buf = (void*) malloc( BUF_SIZE );
  if ( write_buf == 0 || read_buf == 0) {
    return -1;
  }
  xdrmem_create( &xdr_in, read_buf, BUF_SIZE, XDR_DECODE );
  xdrmem_create( &xdr_out, write_buf, BUF_SIZE, XDR_ENCODE );
  return 0;
}

PUBLIC int InitSockets( void ){
  int i;
  char *startenv, serverCmd[ 100 ], *serverEnv;
  long test=charPACK('B','a','c','k');
  int noProcs = *(int*)get_param( CONFIG_noProcs );
  int id = *(int*)get_param( CONFIG_id );
  Process *dem;

  if (getenv("SSTEP")) sstep = 1;
  if (getenv("LINES")) lline = 1;
  if (getenv("XDR")) dxdr = 1;
  if ((startenv=getenv("START"))) start=atoi(startenv);

  socketIn = (Channel *)get_param( CONFIG_socketIn );
  socketOut = (Channel *)get_param( CONFIG_socketOut );

  socketOutLock = SemAlloc( 1 );
  synchChan = (Channel*)malloc( MAX_SOCKET_PROCS * sizeof( Channel ) );
  if ( synchChan == 0 ) return -1;

  for ( i = 0; i < MAX_SOCKET_PROCS; i++ ) {
    ChanInit( &synchChan[ i ] );
  }

  if ( id == 0 ) { /* root transputer*/
    if ( serverEnv = getenv("SSERVER") ) {
      strcpy( serverCmd, serverEnv );
    } else {
      strcpy( serverCmd, "sserver /dev/nap7 > /dev/null 2>&1 &" );
    }
    system( serverCmd );
    atexit( cleanup );
  }

  dem = ProcAlloc( Demuxer, 0, 2, synchChan, socketIn );
  ProcRun( dem );

  if ( RegisterFork_with_SocketLib() == -1 ) {
    fprintf( stderr, "out of memory\n" );
    exit( 1 );
  }

  return 0;
}

/* MACROS */

#define DO(x) { char test[100]; if (dxdr) printf("XPUTER:%s\n",#x); \
if (sstep) gets(test); \
if (x == 0) {printf("XDR failure\n"); return -1;} }

#define FD_TRANS FOPEN_MAX
#define OWN(fd) ((fd) >= FD_TRANS)

PRIVATE int FLUSH( void ){
  char* xdrptr;
  int nbytes;
  xdrptr = xdr_out.x_base;
  nbytes = xdr_getpos( &xdr_out );
  SemWait( socketOutLock );
  ChanOutInt( socketOut, ForkId() );
  ChanOutInt( socketOut, nbytes );
  ChanOut( socketOut, xdrptr, nbytes );
  SemSignal( socketOutLock );
  xdr_setpos( &xdr_out, 0 );
}

/* #define SYNCH() ChanInChar( &synchChan[ ForkId() ] ) */

void SYNCH( void ) {
  int id = ForkId();
  ChanInChar( &synchChan[ id ] );
}

PUBLIC int socket(int domain,int type,int protocol){
  long result,command=charPACK(SERV_socket,domain,type,protocol);
L  DO(xdr_long(&xdr_out,&command));

   FLUSH();

L  printf("socket(%d,%d,%d)=",domain,type,protocol); fflush(stdout);

   SYNCH();
L  DO(xdr_long(&xdr_in,&result));
   SYNCH();

L  printf("%ld\n",result);
L  if (result >= 0) result += FD_TRANS;
L  return (int)result;

}

PUBLIC int listen(int fd,int backlog){
  long result,command;
  fd -= FD_TRANS;
  command=charPACK(SERV_listen,(char)backlog,shortLO(fd),shortHI(fd));
  printf("listen(%d,%d)=",fd,backlog); fflush(stdout);
  DO(xdr_long(&xdr_out,&command));

  FLUSH();

  SYNCH();
  DO(xdr_long(&xdr_in,&result));
  printf("%ld\n",result);
  if (result == -1)
    xdr_long(&xdr_in, (long*)&errno);
  SYNCH();
  return (int)result;
}

PUBLIC int accept(int fd,struct sockaddr *addr,int *addrlen){
  long command,result;
  fd -= FD_TRANS;
/* command */
  command=charPACK(SERV_accept,isNULL(addr),shortLO(fd),shortHI(fd));
  printf("accept(%d,&,&%d)=",fd,*addrlen); fflush(stdout);
L  DO(xdr_long(&xdr_out,&command));

   FLUSH();

/* result */
  SYNCH();
L  DO(xdr_long(&xdr_in,&result));
L  printf("%ld\n",result);
/* if success, value return */
L  if (result != -1) {
L    if (!isNULL(addr)) {
L      DO(xdr_sockaddr(&xdr_in,addr,addrlen));
     }
L    result += FD_TRANS;
  } else
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
  return (int)result;
}

PUBLIC int select(int nfds,fd_set *readfds,fd_set *writefds,
		  fd_set *exceptfds, struct timeval *timeout){
  long result;
  int nlongs,i;
  long command=charPACK(SERV_select,
			bitPACK(isNULL(readfds),
				isNULL(writefds),
				isNULL(exceptfds),
				isNULL(timeout),_,_,_,_),
			shortHI(nfds),
			shortLO(nfds)  );
/* command */
  DO(xdr_long(&xdr_out,&command));
/* parameters, only send fdsets and timeout if non null */
  nlongs = howmany(nfds,NFDBITS);
  if (readfds != 0)
    for (i=0; i<nlongs; i++)
      DO(xdr_long(&xdr_out, &readfds->fds_bits[i]));
  if (writefds != 0)
    for (i=0; i<nlongs; i++)
      DO(xdr_long(&xdr_out, &writefds->fds_bits[i]));
  if (exceptfds != 0)
    for (i=0; i<nlongs; i++)
      DO(xdr_long(&xdr_out, &exceptfds->fds_bits[i]));
  if (timeout != 0) {
    DO(xdr_long(&xdr_out, &timeout->tv_sec));
    DO(xdr_long(&xdr_out, &timeout->tv_usec));
  }

  FLUSH();

/* result */
  SYNCH();
  if (xdr_long(&xdr_in, &result) == 0) return -1;
/* if success, value return */
  if (result > 0){
    if (readfds != 0)
      for (i=0; i<nlongs; i++)
	DO(xdr_long(&xdr_in, &readfds->fds_bits[i]));
    if (writefds != 0)
      for (i=0; i<nlongs; i++)
	DO(xdr_long(&xdr_in, &writefds->fds_bits[i]));
/* if failure, return errno */
  } else if (result == -1) {
    xdr_long(&xdr_in, (long*)&errno);
  }
  SYNCH();
/* or timeout with result = 0, no action */
  return (int)result;
}

PUBLIC int connect(int fd,const struct sockaddr *name,int namelen){
  long result,command;
  fd -= FD_TRANS;
/* command */
  command = charPACK( SERV_connect, isNULL(name), shortLO(fd), shortHI(fd) );
  printf("connect(%d,*,%d)=",fd,namelen); fflush(stdout);
  DO(xdr_long(&xdr_out,&command));
/* if non null, param name */
  if (name != 0)
    DO(xdr_sockaddr(&xdr_out,(struct sockaddr *)name,&namelen));

  FLUSH();

/* result */
  SYNCH();
  DO(xdr_long(&xdr_in,&result));
  printf("%ld\n",result);
/* on failure get errno */
  if (result == -1) 
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
  return (int)result;
}

PUBLIC int bind(int fd,const struct sockaddr *name,int namelen){
  long result,command;
  fd -= FD_TRANS;
/* command */
  command=charPACK(SERV_bind,_,shortLO(fd),shortHI(fd));
  printf("bind(%d,*,%d)=",fd,namelen); fflush(stdout);
  DO(xdr_long(&xdr_out,&command));
/* param name */
  DO(xdr_sockaddr(&xdr_out,(struct sockaddr *)name,&namelen));

  FLUSH();

/* result */
  SYNCH();
  DO(xdr_long(&xdr_in,&result));
  printf("%ld\n",result);
/* on failure get errno */
  if (result == -1) 
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
  return (int)result;
}

PUBLIC int Tread(int fd,char *buf,u_int nbytes){
  long result,command;
  if (!OWN(fd)) return read(fd,buf,nbytes);
  if (nbytes == 0) return 0;
/* command */
  fd -= FD_TRANS;
  printf("read(%d,*,%d)=",fd,nbytes); fflush(stdout);
  command=charPACK(SERV_read,_,shortLO(fd),shortHI(fd));
  DO(xdr_long(&xdr_out,&command));
/* params */
  DO(xdr_int(&xdr_out,(int*)&nbytes));

  FLUSH();

/* result */
  SYNCH();
  DO(xdr_long(&xdr_in,&result));
  printf("%ld\n",result);
/* on success */
  if (result > 0) {
    DO(xdr_opaque(&xdr_in,buf,(u_int)result));
  } else  if (result == -1)
/* on failure get errno */
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
/* return no of bytes read */
  return (int)result;
}

PUBLIC int Twrite(int fd,const char *buf,u_int nbytes){
  long result, command;
  if (!OWN(fd)) return write(fd,buf,nbytes);

/* command */
  fd -= FD_TRANS;
  command=charPACK(SERV_write,_,shortLO(fd),shortHI(fd));
  printf("write(%d,*,%d)=",fd,nbytes); fflush(stdout);
  DO(xdr_long(&xdr_out,&command));
/* params */
  DO(xdr_int(&xdr_out,(int*)&nbytes));
  DO(xdr_opaque(&xdr_out,(char*)buf,nbytes));

  FLUSH();

/* result */
  SYNCH();
  DO(xdr_long(&xdr_in,&result));
  printf("%ld\n",result);
  if (result == -1)
/* on failure get errno */
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
/* return no of bytes read */
  return (int)result;
}

PUBLIC int Tclose(int fd){
  long result,command;
  if (!OWN(fd)) return close(fd);

/* command */
  fd -= FD_TRANS;
  command=charPACK(SERV_close,_,shortLO(fd),shortHI(fd));
  printf("close(%d)=",fd); fflush(stdout);
  DO(xdr_long(&xdr_out,&command));

  FLUSH();

/* result */
  SYNCH();
  DO(xdr_long(&xdr_in,&result));
  printf("%ld\n",result);
/* on failure get errno */
  if (result == -1)
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
  return (int)result;
}

PUBLIC int recv(int fd,char *buf,int len,int flags){
  long result,command;
  if (!OWN(fd)) {
    errno=ENOTSOCK;
    return -1;
  }
  if (len == 0) return 0;
/* command */
  fd -= FD_TRANS;
  command=charPACK(SERV_recv,_,shortLO(fd),shortHI(fd));
  printf("recv(%d,*,%d,%d)=",fd,len,flags); fflush(stdout);
  DO(xdr_long(&xdr_out,&command));
/* params */
  DO(xdr_int(&xdr_out,(int*)&len));
  DO(xdr_int(&xdr_out,(int*)&flags));

  FLUSH();

/* result */
  SYNCH();
  DO(xdr_long(&xdr_in,&result));
  printf("%ld\n",result);
/* on success */
  if (result > 0) {
    DO(xdr_opaque(&xdr_in,buf,(u_int)result));
  } else  if (result == -1)
/* on failure get errno */
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
/* return no of bytes read */
  return (int)result;
}

PUBLIC int recvfrom(int fd,char* buf,int len,int flags,
		    struct sockaddr *from,int *fromlen){
/* command */
  long result,command;
  if (!OWN(fd)) {
    errno=ENOTSOCK;
    return -1;
  }
  if (len == 0) return 0;
  fd -= FD_TRANS;
  command=charPACK(SERV_recvfrom,isNULL(from),shortLO(fd),shortHI(fd));
  printf("recvfrom(%d,*,%d,%d,&*,&%d)=",fd,len,flags,*fromlen); fflush(stdout);
  DO(xdr_long(&xdr_out,&command));
/* params */
  DO(xdr_int(&xdr_out,(int*)&len));
  DO(xdr_int(&xdr_out,(int*)&flags));

  FLUSH();

/* result */
  SYNCH();
  DO(xdr_long(&xdr_in,&result));
  printf("%ld\n",result);
/* on success */
  if (result > 0) {
    if (from != 0) DO(xdr_sockaddr(&xdr_in,from,fromlen));
    DO(xdr_opaque(&xdr_in,buf,(u_int)result));
  } else  if (result == -1)
/* on failure get errno */
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
/* return no of bytes read */
  return (int)result;
}

PUBLIC int send(int fd, char *msg, int len, int flags){
  long result,command;
  if (!OWN(fd)) {
    errno=ENOTSOCK;
    return -1;
  }
/* command */
  fd -= FD_TRANS;
  command=charPACK(SERV_send,_,shortLO(fd),shortHI(fd));
  printf("send(%d,*,%d,%d)=",fd,len,flags); fflush(stdout);
  DO(xdr_long(&xdr_out,&command));
/* params */
  DO(xdr_int(&xdr_out,(int*)&len));
  DO(xdr_int(&xdr_out,(int*)&flags));
  FLUSH();

  SYNCH();
  DO(xdr_opaque(&xdr_in,msg,(u_int)len));

/* result */
  DO(xdr_long(&xdr_in,&result));
  printf("%ld\n",result);
/* on failure get errno */
  if (result == -1)
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
/* return no of bytes sent */
  return (int)result;
}

PUBLIC int sendto(int fd,char *msg,int len,int flags,
		  struct sockaddr *to, int tolen){
  long result,command;
  if (!OWN(fd)) {
    errno=ENOTSOCK;
    return -1;
  }
/* command */
  fd -= FD_TRANS;
  command=charPACK(SERV_sendto,_,shortLO(fd),shortHI(fd));
  printf("sendto(%d,\"%s\",%d,%d,{%d,%s})\n",fd,msg,len,flags,
	 to->sa_family,to->sa_data);
  DO(xdr_long(&xdr_out,&command));
/* params */
  DO(xdr_int(&xdr_out,(int*)&len));
  DO(xdr_int(&xdr_out,(int*)&flags));
  DO(xdr_sockaddr(&xdr_out,to,&tolen));
  DO(xdr_opaque(&xdr_out,msg,(u_int)len));

  FLUSH();

/* result */
  SYNCH();
  DO(xdr_long(&xdr_in,&result));
  printf("%ld\n",result);
/* on failure get errno */
  if (result == -1)
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
/* return no of bytes sent */
  return (int)result;
}

PUBLIC int getsockname(int fd,struct sockaddr *name,int *namelen){
  long command,result;
  fd -= FD_TRANS;
/* command */
  command=charPACK(SERV_getsockname,_,shortLO(fd),shortHI(fd));
  printf("getsockname(%d,&*,&%d)=",fd,*namelen); fflush(stdout);
  DO(xdr_long(&xdr_out,&command));

  FLUSH();

/* result */
  SYNCH();
  DO(xdr_long(&xdr_in,&result));
  printf("%ld\n",result);
/* if success, value return */
  if (result != -1) {
    DO(xdr_sockaddr(&xdr_in,name,namelen));
    result += FD_TRANS;
  } else
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
  return (int)result;
}

PUBLIC struct hostent *gethostbyname(const char *name){
  long command,len;
  static struct hostent entry;
  int res;
/* command */
  len = strlen(name);
  command=charPACK(SERV_gethostbyname,_,shortLO(len),shortHI(len));
  printf("gethostbyname(\"%s\")=",name); fflush(stdout);
L  if (xdr_long(&xdr_out, &command) == 0) return 0;
L  if (xdr_opaque(&xdr_out,(char*)name,(u_int)len) == 0) return 0;

  FLUSH();

  SYNCH();
  res = xdr_hostent(&xdr_in,&entry);
  SYNCH();
  if ( res == 0 ) return 0;
L  printf("*\n"); fflush(stdout);
/*
  printf("entry->h_name=\"%s\"\n",entry.h_name);
  printf("entry->h_aliases=\"%s\"\n",(*entry.h_aliases)?(*entry.h_aliases):"0");
  printf("entry->h_addrtype=%d\n",entry.h_addrtype);
  printf("entry->h_length=%d\n",entry.h_length);
  if (entry.h_addr_list[0]) printf("entry->h_addr_list=%d.%d.%d.%d\n",
				   entry.h_addr_list[0][0],entry.h_addr_list[0][1],
				   entry.h_addr_list[0][2],entry.h_addr_list[0][3]);
  if (entry.h_addr_list[1]) printf("entry->h_addr_list=%d.%d.%d.%d\n",
				   entry.h_addr_list[1][0],entry.h_addr_list[1][1],
				   entry.h_addr_list[1][2],entry.h_addr_list[1][3]);
  printf("end of entry\n");

*/
L  return &entry;
}

int getpeername(int fd, struct sockaddr* name, int* namelen){
  long command,result;
/* command */
  command=charPACK(SERV_getpeername,_,shortLO(fd),shortHI(fd));
  printf("getpeername(%d)=",fd); fflush(stdout);
L  if (xdr_long(&xdr_out, &command) == 0) return 0;

  FLUSH();

/* result */
  SYNCH();
  DO(xdr_long(&xdr_in,&result));
  printf("%ld\n",result);
/* if success, value return */
  if (result != -1) {
    DO(xdr_sockaddr(&xdr_in,name,namelen));
  } else
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
  return (int)result;
}

int shutdown(int fd, int how){
  long command,result;
/* command */
  fd -= FD_TRANS;
  command=charPACK(SERV_shutdown,shortLO(how),shortLO(fd),shortHI(fd));
  printf("shutdown(%d,%d)=",fd,how); fflush(stdout);
L  if (xdr_long(&xdr_out, &command) == 0) return 0;

  FLUSH();

/* result */
  SYNCH();
  DO(xdr_long(&xdr_in,&result));
  printf("%ld\n",result);
  if (result == -1)
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
  return (int)result;
}

int fcntl(int fd, long cmd, long arg){
  long result,command;

/* command */
  fd -= FD_TRANS;
  command=charPACK(SERV_fcntl,_,shortLO(fd),shortHI(fd));
  printf("fcntl(%d,0x%08x,0x%08x)=",fd,cmd,arg); fflush(stdout);
  DO(xdr_long(&xdr_out,&command));
  DO(xdr_long(&xdr_out,&cmd));
  DO(xdr_long(&xdr_out,&arg));

  FLUSH();

/* result */
  SYNCH();
  DO(xdr_long(&xdr_in,&result));
  printf("%ld\n",result);
/* on failure get errno */
  if (result == -1)
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
  return (int)result;
}

int chdir(char* dir){
  long result, command;

/* disable while we're not using socket server */

  return 0;

/* command */
  command=charPACK(SERV_chdir,_,_,_);
  printf("chdir(\"%s\")=",dir);
  DO(xdr_long(&xdr_out,&command));
  DO(xdr_string(&xdr_out,&dir,2048));

  FLUSH();

/* result */
  SYNCH();
  DO(xdr_long(&xdr_in,&result));
  printf("%ld\n",result);
/* on failure get errno */
  if (result == -1)
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
  return (int)result;
}

int ss_puts(char *string){
  long result, command;

/* command */
  command=charPACK(SERV_ss_puts,_,_,_);
  DO(xdr_long(&xdr_out,&command));
  DO(xdr_string(&xdr_out,&string,2048));

  FLUSH();

/* result */
  SYNCH();
  DO(xdr_long(&xdr_in,&result));
/*  printf("%ld\n",result); dont do that !!! hangs */
/* on failure get errno */
  if (result == -1)
    xdr_long(&xdr_in,(long*)&errno);
  SYNCH();
  return (int)result;
}
