/**********************************************************************\
*                                                                      *
*                  Copyright (c) Exeter University,		       *
*                  Department of Computer Science,                     *
*                 (aba@uk.ac.ex.dcs) Adam Back, 1994                   *
*                                                                      *
*  GNU GPL (General Public License) Applies.                           *
*                                                                      *
\**********************************************************************/
#include <string.h>
#include <stdlib.h>
#include "t_sockxdr.h"
#include "t_xdr.h"

#ifdef sun
/* SPARC */
#include <sys/types.h>
#include <rpc/rpc.h>
#include <netdb.h>
#else
/* TRANSPUTER */
#define TRANSPUTER
#include "t_socket.h"
#endif


/* required XDR functions */

PUBLIC bool_t xdr_sockaddr(XDR *xdrs, struct sockaddr *name, int *namelen){
  long templong;
  switch (xdrs->x_op){
  case XDR_ENCODE:
    templong = shortPACK( (short)*namelen, name->sa_family );
    if (xdr_long(xdrs, &templong) == 0) return 0;
    if (xdr_opaque(xdrs,(char*)name->sa_data,*namelen - 2) == 0) return 0;
    return 1;
  case XDR_DECODE:
    if (xdr_long(xdrs,&templong) == 0) return 0;
    *namelen = SHORT(1,templong);
    name->sa_family=SHORT(2,templong);
#ifdef TRANSPUTER
#ifdef DEBUG
    printf("xdr_sockaddr:\n");
    printf("namelen=%d, family=%d\n",*namelen,name->sa_family);
#endif
#endif
    if (xdr_opaque(xdrs,name->sa_data,*namelen - 2) == 0) return 0;
    return 1;
  case XDR_FREE:;
  default:;
  }
  return 0;
}

PUBLIC bool_t xdr_hostent(XDR *xdrs, struct hostent *entry){
  long n,i;
  char **list;
  switch (xdrs->x_op){
  case XDR_ENCODE:
    if (xdr_string(xdrs, &(entry->h_name), strlen(entry->h_name)) == 0) return 0;
    for (n=0,list=entry->h_aliases; *list; list++,n++);
    if (xdr_int(xdrs, (int*)&n) == 0) return 0;
    for (list=entry->h_aliases; *list; list++){
      if (xdr_string(xdrs, list, strlen(*list)) == 0) return 0;
    }
    if (xdr_int(xdrs, (int*)&(entry->h_addrtype)) == 0) return 0;
    if (xdr_int(xdrs, (int*)&(entry->h_length)) == 0) return 0;
    for (n=0,list=entry->h_addr_list; *list; list++,n++);
    if (xdr_int(xdrs, (int*)&n) == 0) return 0;
    for (list=entry->h_addr_list; *list; list++){
      if (xdr_opaque(xdrs, *list,entry->h_length) == 0)
	return 0;
    }
    return 1;
  case XDR_DECODE:
    if (xdr_string(xdrs, &(entry->h_name), 256) == 0) return 0;
    if (xdr_int(xdrs, (int*)&n) == 0) return 0;
    entry->h_aliases = (char**)malloc(sizeof(char**)*(n+1));
    for (list=entry->h_aliases,i=0; i<n; i++,list++){
      *list = 0;                             /* let xdr_string do the malloc */
      if (xdr_string(xdrs, list, 256) == 0) return 0;
    }
    *list = 0;                               /* put a 0 on the end */
    if (xdr_int(xdrs, (int*)&(entry->h_addrtype)) == 0) return 0;
    if (xdr_int(xdrs, (int*)&(entry->h_length)) == 0) return 0;

    if (xdr_int(xdrs, (int*)&n) == 0) return 0;
    entry->h_addr_list = (char**)malloc(sizeof(char**)*(n+1));
    for (list=entry->h_addr_list,i=0; i<n; i++,list++){
      *list = (char*)malloc(entry->h_length);
      if (xdr_opaque(xdrs, *list, entry->h_length) == 0)
	return 0;
    }
    *list = 0;                               /* put a 0 on the end */
    return 1;
  case XDR_FREE:
  default:;
  }
  return 0;
}
