/**********************************************************************\
*                                                                      *
*                  Copyright (c) Exeter University,		       *
*                  Department of Computer Science,                     *
*                 (aba@uk.ac.ex.dcs) Adam Back, 1994                   *
*                                                                      *
*  GNU GPL (General Public License) Applies.                           *
*                                                                      *
\**********************************************************************/
/*
 *  Transputer XDR
 *
 *  Implements the PUBLIC functions declared in xdr.h
 *
 */
#include <string.h>
#include <stdlib.h>
#include "t_xdr.h"
#include "t_macros.h"

bool_t t_xdr_short(XDR *xdrs, short *sp){
  long temp;
  if (sp==0) return 0;
  temp=*sp;
  switch (xdrs->x_op){
  case XDR_DECODE:
    if (xdrs->x_ops->x_getlong(xdrs,&temp) == 0) return 0;
    *sp=(short)temp;
    return 1;
  case XDR_ENCODE:
    return xdrs->x_ops->x_putlong(xdrs,&temp);
  default:;
  }
  return 0;
}

bool_t t_xdr_int(XDR *xdrs,int *ip){
L  return xdr_long(xdrs,(long*)ip);
}

bool_t t_xdr_long(XDR *xdrs, long *lp){
L  switch (xdrs->x_op){
  case XDR_DECODE:
L    return xdrs->x_ops->x_getlong(xdrs,lp);
  case XDR_ENCODE:
L    return xdrs->x_ops->x_putlong(xdrs,lp);
  default:;
L  }
L  return 0;
}

bool_t t_xdr_array(XDR *xdrs, caddr_t *arrp, u_int *sizep,
		 u_int maxsize, u_int elsize, xdrproc_t elproc){
  int i;
  caddr_t ptr;

  switch (xdrs->x_op){
  case XDR_DECODE:
    if (arrp == 0) return 0;
    if (xdrs->x_ops->x_getlong(xdrs,(long*)sizep) == 0)
      return 0;
    if (*sizep > maxsize) return 0;
    if (*arrp == 0)
      if ((*arrp = malloc(*sizep * elsize)) == 0)
	return 0;
    for (i = 0, ptr = *arrp; i < *sizep; i++, ptr += elsize)
      if (elproc(xdrs, ptr) == 0)
	return 0;
    return 1;
  case XDR_ENCODE:
    if (sizep == 0 || *sizep > maxsize) return 0;
    if (xdrs->x_ops->x_putlong(xdrs,(long*)sizep)==0)
      return 0;
    for (i = 0, ptr = *arrp; i < *sizep; i++, ptr += elsize)
      if (elproc(xdrs, ptr)==0)
	return 0;
    return 1;
  default:;
  }
  return 0;
}

bool_t t_xdr_float(XDR *xdrs, float *fp){
  switch (xdrs->x_op){
  case XDR_DECODE:
    return xdrs->x_ops->x_getlong(xdrs,(long*)fp);
  case XDR_ENCODE:
    return xdrs->x_ops->x_putlong(xdrs,(long*)fp);
  default:;
  }
}

bool_t t_xdr_double(XDR *xdrs, double *dp){
  if (dp == 0) return 0;
  switch (xdrs->x_op){
  case XDR_DECODE:
    return xdrs->x_ops->x_getlong(xdrs,(long*)dp+1) &&
           xdrs->x_ops->x_getlong(xdrs,(long*)dp);
  case XDR_ENCODE:
    return xdrs->x_ops->x_putlong(xdrs,(long*)dp+1) &&
           xdrs->x_ops->x_putlong(xdrs,(long*)dp);
  default:;
  }
  return 0;
}

bool_t t_xdr_string(XDR *xdrs, char **string, u_int maxlen){
  u_int temp;
  if (string == 0) return 0;
  switch (xdrs->x_op){
  case XDR_ENCODE:
    temp = strlen(*string);
    if (temp > maxlen) return 0;
    return xdrs->x_ops->x_putlong(xdrs,(long*)&temp) &&
           xdrs->x_ops->x_putbytes(xdrs,*string,temp);
  case XDR_DECODE:
    if (xdrs->x_ops->x_getlong(xdrs,(long*)&temp) == 0)
      return 0;
    if (temp > maxlen) return 0;
    if (*string == 0)
      if ((*string = (char*)malloc((temp + 1)*sizeof(char))) == 0)
	return 0;
    if (xdrs->x_ops->x_getbytes(xdrs,*string,temp) == 0)
      return 0;
    (*string)[temp] = '\0';
    return 1;
  default:;
  }
  return 0;
}

bool_t t_xdr_bytes(XDR *xdrs, caddr_t *arrp, u_int *sizep, u_int maxsize){
  if (sizep == 0 || arrp == 0) return 0;
  switch (xdrs->x_op){
  case XDR_ENCODE:
    if (*sizep > maxsize) return 0;
    if (xdrs->x_ops->x_putlong(xdrs,(long*)sizep) == 0) return 0;
    return xdrs->x_ops->x_putbytes(xdrs,*arrp,*sizep);
  case XDR_DECODE:
    if (xdrs->x_ops->x_getlong(xdrs,(long*)sizep) == 0) return 0;
    if (*sizep > maxsize) return 0;
    if (*arrp == 0) 
      if ((*arrp = (caddr_t)malloc(*sizep * sizeof(char))) == 0)
	return 0;
    return xdrs->x_ops->x_getbytes(xdrs,*arrp,*sizep);
  default:;
  }
  return 0;
}

bool_t t_xdr_opaque(XDR *xdrs, char *data, u_int bytes){
  if (data == 0) return 0;
  switch (xdrs->x_op){
  case XDR_ENCODE:
    return xdrs->x_ops->x_putbytes(xdrs,data,bytes);
  case XDR_DECODE:
    return xdrs->x_ops->x_getbytes(xdrs,data,bytes);
  default:;
  }
  return 0;
}

/************************************************************************
 *
 * XDR set of internal functions    ( STREAM SET )
 *
 * All internal functions are accessed through the x_ops field
 *
 * The struct stream_ops is declared for the purpose of initialising
 * this pointer
 *
 ***********************************************************************/

PRIVATE bool_t	s_getlong  (XDR *, long *);
PRIVATE bool_t	s_putlong  (XDR *, long *);
PRIVATE bool_t	s_getbytes (XDR *, void *, u_int);
PRIVATE bool_t	s_putbytes (XDR *, void *, u_int);
PRIVATE u_int	s_getpostn (XDR *);
PRIVATE bool_t	s_setpostn (XDR *, u_int);
PRIVATE long *	s_inline   (XDR *, int);
PRIVATE void	s_destroy  (XDR *);

PRIVATE struct xdr_ops stream_ops={
  s_getlong, s_putlong, s_getbytes, s_putbytes,
  s_getpostn, s_setpostn, s_inline, s_destroy
};

#define XDR_FD ((FILE*)(xdrs->x_private))
#define XDR_FBASE ((int)(xdrs->x_base))

#ifndef _ICC
#define SEEK 1
/*#define SEEK (lseek(XDR_FD,0,0) != -1)*/
#else
#define SEEK 1
#endif

PRIVATE char pad[3]={0,0,0};
PRIVATE char junk[3];

#ifndef _ICC
#define fread dev_read

int dev_read(void *buf, int size, int n, FILE *s){
  int i,left,rd,code;
  char *pos=buf;
  for (i = 0; i < n; i++){
    left = size;
    rd = 0;
    while (rd < left){
      code = read(fileno(s),pos+size-left,left);
      if (code < 0) {
	perror("dev_read");
	exit(0);
	return i - 1;
      }
      rd += code;
      left -= code;
    }
    pos += size;
  }
  return n;
}

#define fwrite dev_write
int dev_write(void *buf, int size, int n, FILE *s){
  int i,left,wr,code;
  char *pos=buf;
  for (i = 0; i < n; i++){
    left = size;
    wr = 0;
    while (wr < left){
      code = write(fileno(s),pos+size-left,left);
      if (code < 0) {
	perror("dev_write");
	exit(0);
	return i - 1;
      }
      wr += code;
      left -= code;
    }
    pos += size;
  }
  return n;
}

#endif

PRIVATE bool_t s_getlong(XDR *xdrs, long *lp){
  if (lp == 0) 
    {printf("!!!\n"); return 0;}
  if (fread(lp,4,1,XDR_FD) == 0 && SEEK)
    return 0;
  *lp=ntohl(*lp);
  return 1;
}

PRIVATE bool_t s_putlong(XDR *xdrs, long *lp){
  long l;
  if (lp == 0) 
    {printf("!!!\n"); return 0;}
  l=htonl(*lp);
  return fwrite(&l,4,1,XDR_FD) && SEEK;
}

PRIVATE bool_t s_getbytes(XDR *xdrs, void *buf, u_int bytecount){
  u_int padbytes = ((bytecount + 3) & ~3) - bytecount;
  if (buf == 0)
    {printf("!!!\n"); return 0;}
  return fread(buf,bytecount,1,XDR_FD) && 
    (padbytes == 0 || fread(junk,padbytes,1,XDR_FD)) && SEEK;
}

PRIVATE bool_t s_putbytes(XDR *xdrs, void * buf, u_int bytecount){
  u_int padbytes = ((bytecount + 3) & ~3) - bytecount;
  if (buf == 0)
    {printf("!!!\n"); return 0;}
  return fwrite(buf,bytecount,1,XDR_FD) && 
    (padbytes == 0 || fwrite(pad,padbytes,1,XDR_FD)) && SEEK;
}

PRIVATE u_int s_getpostn(XDR *xdrs){
  fprintf(stderr,"x_getpostn unavailable for stream XDRs\n");
  exit(1);
}

PRIVATE bool_t s_setpostn(XDR *xdrs, u_int posn){
  fprintf(stderr,"x_setpostn unavailable for stream XDRs\n");
  exit(1);
}

PRIVATE long *s_inline(XDR *xdrs, int nbytes ){
  fprintf(stderr,"Internal x_inline unavailable for stream XDRs\n");
  exit(1);
}

PRIVATE void s_destroy(XDR *xdrs){
  fflush(XDR_FD);
}

PUBLIC void t_xdrstdio_create(XDR *xdrs, FILE *fp, enum xdr_op op){
  xdrs->x_op=op;
  xdrs->x_ops=&stream_ops;
  xdrs->x_base=(caddr_t)ftell(fp);
  xdrs->x_private=(caddr_t)fp;
}

#ifdef _ICC
/************************************************************************
 *
 * XDR set of internal functions    ( CHANNEL SET )
 *
 * All internal functions are accessed through the x_ops field
 *
 * The struct channel_ops is declared for the purpose of initialising
 * this pointer
 *
 ***********************************************************************/

PRIVATE bool_t	c_getlong  (XDR *, long *);
PRIVATE bool_t	c_putlong  (XDR *, long *);
PRIVATE bool_t	c_getbytes (XDR *, void *, u_int);
PRIVATE bool_t	c_putbytes (XDR *, void *, u_int);
PRIVATE u_int	c_getpostn (XDR *);
PRIVATE bool_t	c_setpostn (XDR *, u_int);
PRIVATE long *	c_inline   (XDR *, int);
PRIVATE void	c_destroy  (XDR *);

PRIVATE struct xdr_ops channel_ops={
  c_getlong, c_putlong, c_getbytes, c_putbytes,
  c_getpostn, c_setpostn, c_inline, c_destroy
};

#define XDR_CHD ((Channel*)(xdrs->x_private))

PRIVATE bool_t c_getlong(XDR *xdrs, long *lp){
L  if (lp == 0) 
    {printf("!!!\n"); return 0;}
L  *lp = ChanInInt( XDR_CHD );
L  return 1;
}

PRIVATE bool_t c_putlong(XDR *xdrs, long *lp){
L  if (lp == 0) 
    {printf("!!!\n"); return 0;}
L  ChanOutInt( XDR_CHD, *lp );
L  return 1;
}

PRIVATE bool_t c_getbytes(XDR *xdrs, void *buf, u_int bytecount){
  u_int padbytes = ((bytecount + 3) & ~3) - bytecount;
L  if (buf == 0)
    {printf("!!!\n"); return 0;}
L  ChanIn( XDR_CHD, buf, bytecount );
L  if ( padbytes > 0 ) ChanIn( XDR_CHD, junk, padbytes );
L  return 1;
}

PRIVATE bool_t c_putbytes(XDR *xdrs, void * buf, u_int bytecount){
  u_int padbytes = ((bytecount + 3) & ~3) - bytecount;
L  if (buf == 0)
    {printf("!!!\n"); return 0;}

L  ChanOut( XDR_CHD, buf, bytecount );
L  if ( padbytes > 0 ) ChanOut( XDR_CHD, junk, padbytes );
L  return 1;
}

PRIVATE u_int c_getpostn(XDR *xdrs){
  return (u_int)(ftell(XDR_FD) - XDR_FBASE);
}

PRIVATE bool_t c_setpostn(XDR *xdrs, u_int posn){
  return (!fseek(XDR_FD, (long)posn + XDR_FBASE, 0)==-1);
}

PRIVATE long *c_inline(XDR *xdrs, int nbytes ){
  fprintf(stderr,"Internal x_inline unavailable for stream XDRs\n");
  exit(1);
}

PRIVATE void c_destroy(XDR *xdrs){
}

PUBLIC void t_xdrchan_create(XDR *xdrs, Channel *ch, enum xdr_op op){
  xdrs->x_op=op;
  xdrs->x_ops=&channel_ops;
  xdrs->x_private = (caddr_t)ch;
}
#endif
/************************************************************************
 *
 * XDR set of internal functions    ( MEMORY SET )
 *
 * All internal functions are accessed through the x_ops field
 *
 * The struct memory_ops is declared for the purpose of initialising
 * this pointer
 *
 ***********************************************************************/

PRIVATE bool_t	m_getlong  (XDR *, long *);
PRIVATE bool_t	m_putlong  (XDR *, long *);
PRIVATE bool_t	m_getbytes (XDR *, void *, u_int);
PRIVATE bool_t	m_putbytes (XDR *, void *, u_int);
PRIVATE u_int	m_getpostn (XDR *);
PRIVATE bool_t	m_setpostn (XDR *, u_int);
PRIVATE long *	m_inline   (XDR *, int);
PRIVATE void	m_destroy  (XDR *);

PRIVATE struct xdr_ops memory_ops={
  m_getlong, m_putlong, m_getbytes, m_putbytes,
  m_getpostn, m_setpostn, m_inline, m_destroy
};

#define XDR_MEM (xdrs->x_private)
#define XDR_MBASE (xdrs->x_base)
#define XDR_MSIZE (xdrs->x_handy)

PRIVATE bool_t m_getlong(XDR *xdrs, long *lp){
  if (lp == 0) return 0;
  if (XDR_MEM + 4 > XDR_MBASE + XDR_MSIZE)
    return 0;
  memcpy(lp,XDR_MEM,4);
  XDR_MEM += 4;
  *lp=ntohl(*lp);
  return 1;
}

PRIVATE bool_t m_putlong(XDR *xdrs, long *lp){
  long network_long;
  if (lp == 0)
    return 0;
  network_long=htonl(*lp);
  if (XDR_MEM + 4 > XDR_MBASE + XDR_MSIZE)
    return 0;
  memcpy(XDR_MEM, &network_long, 4);
  XDR_MEM += 4;
  return 1;
}

PRIVATE bool_t m_getbytes(XDR *xdrs, void *buf, u_int bytecount){
  if (XDR_MEM + bytecount > XDR_MBASE + XDR_MSIZE)
    return 0;
  memcpy(buf,XDR_MEM,bytecount);
  XDR_MEM += (bytecount + 3) & ~3;
  return 1;
}

PRIVATE bool_t m_putbytes(XDR *xdrs, void * buf, u_int bytecount){
  u_int padbytes = ((bytecount + 3) & ~3) - bytecount;
  if (XDR_MEM + bytecount > XDR_MBASE + XDR_MSIZE)
    return 0;
  memcpy(XDR_MEM,buf,bytecount);
  XDR_MEM += bytecount;
  if (padbytes) {
    memset(XDR_MEM,0,padbytes);
    XDR_MEM += padbytes;
  }
  return 1;
}

PRIVATE u_int m_getpostn(XDR *xdrs){
  return XDR_MEM - XDR_MBASE;
}

PRIVATE bool_t m_setpostn(XDR *xdrs, u_int posn){
  if (posn > XDR_MSIZE) 
    return 0;
  XDR_MEM = XDR_MBASE + posn;
  return 1;
}

PRIVATE long *m_inline(XDR *xdrs, int nbytes){
  long *val;
  if (XDR_MEM + nbytes > XDR_MBASE + XDR_MSIZE)
    return 0;
  val=(long*)XDR_MEM;
  XDR_MEM += nbytes;
  return val;
}

PRIVATE void m_destroy(XDR *xdrs){
}

PUBLIC void t_xdrmem_create(XDR *xdrs,void *addr,u_int size,enum xdr_op op){
  xdrs->x_op=op;
  xdrs->x_ops=&memory_ops;
  xdrs->x_private = xdrs->x_base = (caddr_t)addr;
  xdrs->x_handy=size;
}

/*****************************************************************
 * CONVERSION PRIMITIVES for
 *
 * network -> host
 * host    -> network
 *
 * on longs (32 bits)
 * and shorts (16 bits)
 *
 *****************************************************************/

#ifdef _ICC
PUBLIC short htons(short s){
  char *p=(char*)&s;
  return p[0] << 8 | p[1];
}

PUBLIC long htonl(long l){
  char *p=(char*)&l,ans[4];
  int *i = (int*)ans;
  ans[0] = p[3]; 
  ans[1] = p[2]; 
  ans[2] = p[1]; 
  ans[3] = p[0];
  return *i;
}
#endif
